package nl.pim16aap2.armoredElytra.nbtEditor.ArmoredElytraBuilder;

import nl.pim16aap2.armoredElytra.ArmoredElytra;
import nl.pim16aap2.armoredElytra.nbtEditor.DurabilityManager;
import nl.pim16aap2.armoredElytra.nbtEditor.NBTEditor;
import nl.pim16aap2.armoredElytra.util.ArmorTier;
import nl.pim16aap2.armoredElytra.util.ConfigLoader;
import nl.pim16aap2.armoredElytra.util.EnchantmentContainer;
import org.bukkit.inventory.ItemStack;

import javax.annotation.Nullable;

@SuppressWarnings({"unused", "UnusedReturnValue"})
public class ArmoredElytraBuilder {
    private final NBTEditor nbtEditor;
    private final DurabilityManager durabilityManager;
    private final ConfigLoader config;
    private final ArmoredElytra plugin;

    public ArmoredElytraBuilder(NBTEditor nbtEditor, DurabilityManager durabilityManager,
                                ConfigLoader config, ArmoredElytra plugin) {

        this.nbtEditor = nbtEditor;
        this.durabilityManager = durabilityManager;
        this.config = config;
        this.plugin = plugin;
    }

    /**
     * Creates a new builder for an armored elytra.
     *
     * @return The first step of the new builder.
     */
    public ItemCreator newBuilder() {
        return new Builder(nbtEditor, durabilityManager, config, plugin);
    }

    /**
     * Shortcut for repairing an armored elytra.
     *
     * @param armoredElytra The armored elytra to repair.
     * @param repairItems   The repair item(s) to use for repairing the armored elytra. It is assumed that they are of
     *                      the correct type.
     * @param name          The new name of the output armored elytra. When this is null, {@link
     *                      ArmoredElytra#getArmoredElytraName(ArmorTier)} is used to set the name.
     * @return The new armored elytra.
     */
    public @Nullable ItemStack repair(ItemStack armoredElytra, ItemStack repairItems, @Nullable String name) {
        return newBuilder().ofElytra(armoredElytra).repair(repairItems.getAmount()).withName(name).build();
    }

    /**
     * Shortcut for enchanting an armored elytra.
     *
     * @param armoredElytra The armored elytra to repair.
     * @param sourceItem    The source item from which to copy the enchantments from.
     * @param name          The new name of the output armored elytra. When this is null, {@link
     *                      ArmoredElytra#getArmoredElytraName(ArmorTier)} is used to set the name.
     * @return The new armored elytra.
     */
    public @Nullable ItemStack enchant(ItemStack armoredElytra, ItemStack sourceItem, @Nullable String name) {
        final EnchantmentContainer enchantments = EnchantmentContainer.getEnchantmentsOf(sourceItem, plugin);
        if (enchantments.isEmpty()) {
            return null;
        }
        return newBuilder().ofElytra(armoredElytra).addEnchantments(enchantments).withName(name).build();
    }

    /**
     * Shortcut for creating a new armored elytra from two items.
     *
     * @param elytra    The input item. This should be an (armored) elytra.
     * @param combiner  The item to combine with the elytra. This should either be an armored elytra of the same
     *                  non-NONE tier as the input elytra or a chestplate.
     * @param armorTier The armor tier of the input item. If this is not known, use {@link #combine(ItemStack,
     *                  ItemStack, String)} instead.
     * @param name      The new name of the output armored elytra. When this is null, {@link
     *                  ArmoredElytra#getArmoredElytraName(ArmorTier)} is used to set the name.
     * @return The new armored elytra.
     */
    public ItemStack combine(ItemStack elytra, ItemStack combiner, ArmorTier armorTier, @Nullable String name) {
        return newBuilder().ofElytra(elytra).combineWith(combiner, armorTier).withName(name).build();
    }

    /**
     * See {@link #combine(ItemStack, ItemStack, ArmorTier, String)} for unknown armor tiers.
     */
    public ItemStack combine(ItemStack elytra, ItemStack combiner, @Nullable String name) {
        return newBuilder().ofElytra(elytra).combineWith(combiner).withName(name).build();
    }

    /**
     * Creates a new armored elytra of a specific tier.
     *
     * @param armorTier The tier of the new armored elytra.
     * @return The new armored elytra.
     */
    public ItemStack newArmoredElytra(ArmorTier armorTier) {
        return newBuilder().newItem(armorTier).build();
    }

}
