package nl.pim16aap2.armoredElytra.nbtEditor.ArmoredElytraBuilder;

import nl.pim16aap2.armoredElytra.ArmoredElytra;
import nl.pim16aap2.armoredElytra.nbtEditor.DurabilityManager;
import nl.pim16aap2.armoredElytra.nbtEditor.NBTEditor;
import nl.pim16aap2.armoredElytra.util.ArmorTier;
import nl.pim16aap2.armoredElytra.util.ConfigLoader;
import nl.pim16aap2.armoredElytra.util.EnchantmentContainer;
import nl.pim16aap2.armoredElytra.util.Util;
import org.bukkit.Bukkit;
import org.bukkit.Color;
import org.bukkit.Material;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.LeatherArmorMeta;

import javax.annotation.Nullable;
import java.util.List;

import static nl.pim16aap2.armoredElytra.util.Util.getTextureIdFromArmorTier;

final class Builder implements ItemCreator, ItemCombinator, ItemMetadataSetter {

    private static final Color DEFAULT_LEATHER_COLOR = Bukkit.getServer().getItemFactory().getDefaultLeatherColor();

    private final NBTEditor nbtEditor;
    private final DurabilityManager durabilityManager;
    private final ConfigLoader config;
    private final ArmoredElytra plugin;

    /**
     * The new armored elytra that will be returned at the end of the build process.
     */
    private ItemStack newArmoredElytra;
    /**
     * The combined enchantments of the input items.
     */
    private EnchantmentContainer combinedEnchantments;
    /**
     * The current armor tier of the input elytra.
     */
    private ArmorTier currentArmorTier;
    /**
     * The durability of the output armored elytra.
     */
    private int durability;


    /**
     * The armor tier of the output armored elytra. This defaults to {@link #currentArmorTier} if this isn't set.
     */
    private @Nullable ArmorTier newArmorTier;
    /**
     * The name of the output armored elytra. This defaults to {@link ArmoredElytra#getArmoredElytraName(ArmorTier)}
     * when not overridden.
     */
    private @Nullable String name;
    /**
     * The lore of the output armored elytra. This defaults to {@link ArmoredElytra#getElytraLore(ArmorTier)} when
     * not overridden.
     */
    private @Nullable List<String> lore;
    /**
     * The color of the output armored elytra. By default, the existing color (if any is used). When combined with
     * another item, the color is inferred using {@link #getItemColor(ItemStack, ItemStack)}.
     */
    private @Nullable Color color;
    /**
     * Whether the output armored elytra should be unbreakable. This defaults to {@link ConfigLoader#unbreakable()}
     * when not overridden.
     */
    private @Nullable Boolean isUnbreakable = null;
    /**
     * The texture id (custom_model_data) to be applied to the item. If null, no custom texture id will be set.
     */
    private @Nullable Integer textureId = null;

    Builder(NBTEditor nbtEditor, DurabilityManager durabilityManager,
            ConfigLoader config, ArmoredElytra plugin) {
        this.nbtEditor = nbtEditor;
        this.durabilityManager = durabilityManager;
        this.config = config;
        this.plugin = plugin;
    }

    @Override
    public ItemStack build() {
        // Get default values if unset.
        newArmorTier = newArmorTier == null ? currentArmorTier : newArmorTier;
        name = name == null ? plugin.getArmoredElytraName(newArmorTier) : name;
        lore = lore == null ? plugin.getElytraLore(newArmorTier) : lore;

        isUnbreakable = isUnbreakable == null ? config.unbreakable() : isUnbreakable;

        final ItemStack output = nbtEditor.addArmorNBTTags(newArmoredElytra, newArmorTier,
                isUnbreakable, name, lore, color, textureId);
        durabilityManager.setDurability(output, durability, newArmorTier);
        combinedEnchantments.applyEnchantments(output);

        return output;
    }

    @Override
    public ItemMetadataSetter withName(@Nullable String name) {
        this.name = name;
        return this;
    }

    @Override
    public ItemMetadataSetter withColor(@Nullable Color color) {
        this.color = color;
        return this;
    }

    @Override
    public ItemMetadataSetter withTexture(@Nullable Integer textureId) {
        if (textureId != null) {
            this.textureId = textureId;
        } else if (newArmorTier != null) {
            this.textureId = getTextureIdFromArmorTier(newArmorTier);
        } else {
            this.textureId = null;
        }
        return this;
    }

    @Override
    public ItemMetadataSetter withLore(@Nullable List<String> lore) {
        this.lore = lore;
        return this;
    }

    @Override
    public ItemMetadataSetter unbreakable(boolean isUnbreakable) {
        this.isUnbreakable = isUnbreakable;
        return this;
    }

    @Override
    public ItemMetadataSetter repair(int count) {
        if (currentArmorTier == ArmorTier.NONE) {
            throw new IllegalArgumentException("Non-armored elytras cannot be repaired!");
        }
        durability = durabilityManager.getRepairedDurability(newArmoredElytra, count, currentArmorTier);
        return this;
    }

    @Override
    public ItemMetadataSetter addEnchantments(EnchantmentContainer enchantmentContainer) {
        combinedEnchantments.merge(enchantmentContainer);
        return this;
    }

    @Override
    public ItemMetadataSetter addEnchantments(ItemStack sourceItem) {
        return addEnchantments(EnchantmentContainer.getEnchantmentsOf(sourceItem, plugin));
    }

    @Override
    public ItemMetadataSetter combineWith(ItemStack item, ArmorTier armorTier) {
        // Note: This method is almost always run. Important building steps should happen here.

        if (armorTier == ArmorTier.NONE && !Util.isChestPlate(item)) {
            throw new IllegalArgumentException("Non-armored elytras can only be combined with chest plates!");
        }

        newArmorTier = armorTier;
        if (currentArmorTier == ArmorTier.NONE &&
                item.getType().equals(Material.ELYTRA) && newArmorTier != ArmorTier.NONE) {
            throw new IllegalArgumentException("A regular elytra cannot be combined with an armored one!");
        }

        withColor(getItemColor(newArmoredElytra, item)).withTexture(null);

        addEnchantments(item);

        durability = durabilityManager.getCombinedDurability(newArmoredElytra, item,
                currentArmorTier, newArmorTier);
        return this;
    }

    @Override
    public ItemMetadataSetter combineWith(ItemStack item) {
        final ArmorTier armorTier = item.getType().equals(Material.ELYTRA) ?
                nbtEditor.getArmorTier(item) : Util.armorToTier(item.getType());
        return combineWith(item, armorTier);
    }

    @Override
    public ItemMetadataSetter upgradeToTier(ArmorTier armorTier) {
        newArmorTier = armorTier;
        return this;
    }

    @Override
    public ItemCombinator ofElytra(ItemStack elytra) {
        if (!elytra.getType().equals(Material.ELYTRA)) {
            throw new IllegalArgumentException("Expected elytra as input, but got: " + elytra);
        }

        newArmoredElytra = new ItemStack(elytra);

        if (currentArmorTier == null) {
            currentArmorTier = nbtEditor.getArmorTier(elytra);
        }

        combinedEnchantments = EnchantmentContainer.getEnchantmentsOf(newArmoredElytra, plugin);

        durability = durabilityManager.getRealDurability(newArmoredElytra, currentArmorTier);
        return this;
    }

    @Override
    public ItemMetadataSetter newItem(ArmorTier armorTier) {
        currentArmorTier = newArmorTier = armorTier;
        ofElytra(new ItemStack(Material.ELYTRA));
        return this;
    }

    /**
     * Gets the color of the item if the item has a color.
     * <p>
     * See {@link LeatherArmorMeta#getColor()}.
     *
     * @param itemA The first {@link ItemStack} to check.
     * @param itemB The second {@link ItemStack} to check.
     * @return The color of the item, if it has a color, otherwise null.
     */
    private @Nullable Color getItemColor(final ItemStack itemA, final ItemStack itemB) {
        final @Nullable Color colorA = getItemColor(itemA);
        if (colorA != null && !colorA.equals(DEFAULT_LEATHER_COLOR)) {
            return colorA;
        }

        final @Nullable Color colorB = getItemColor(itemB);
        return colorB != null ? colorB : colorA;
    }

    /**
     * Gets the colors of an item if available.
     * <p>
     * This currently only applies to leather armor(ed elytras).
     *
     * @param itemStack The item to analyze.
     * @return The color of the item, if available, otherwise null.
     */
    private @Nullable Color getItemColor(final ItemStack itemStack) {
        if (itemStack.getType() == Material.ELYTRA) {
            return nbtEditor.getColorOfArmoredElytra(itemStack);
        }

        if (!itemStack.hasItemMeta() || !(itemStack.getItemMeta() instanceof LeatherArmorMeta)) {
            return null;
        }

        return ((LeatherArmorMeta) itemStack.getItemMeta()).getColor();
    }

}
